import ij.*;
import ij.plugin.*;
import ij.gui.*;
import ij.process.*;



/****************************************************************************************
 FOCUS-SIS.java
*****************************************************************************************
Nicolas BOUSSION
nboussion@yahoo.fr
February 2004

Version 2

This ImageJ plugin normalizes and  subtracts two SPECT (nuclear medicine) images
to help locating epileptogenic foci in partial seizures.

A peri-ictal and an inter-ictal exams are needed. They are supposed to be well 
co-registered. A GUI allows the user to choose between relative or absolute subtraction
(Lassen's correction).

The normalization method is described in :
Boussion N, Houzard C, Ostrowsky K, Ryvlin P, Mauguire F, Cinotti L. Automated 
detection of local normalization areas for ictal-interictal subtraction brain SPECT
imaging. J Nucl Med 2002; 43: 1419-1425.

Please cite this paper in any publication concerning results obtained from this plugin.

NEW : This version now supports 8-bit, 16-bit and 32-bit images
 
*****************************************************************************************/ 

public class FOCUS_SIS implements PlugIn {

    static final int RELATIVE=0, LASSEN=1, ECD=1, HMPAO=0;
    static String[] subt = {"Relative", "LASSEN (absolute)"};
    static String[] mol = {"HMPAO","ECD","Default"};
    static int subtract = LASSEN;
    static int molecule = ECD;
    static double alpha=2.6, MEAN1, MEAN2;
    int[] wList;
    private String[] titles;
    int i1Index;
    int i2Index;
    ImagePlus i1;
    ImagePlus i2;
    ImagePlus cMask;
    

    public void run(String arg) {
        
        wList = WindowManager.getIDList();
        if (wList==null || wList.length<2) {
            IJ.showMessage("FOCUS-SIS", "Two open windows are required");
            return;
        }
        titles = new String[wList.length];
        for (int i=0; i<wList.length; i++) {
            ImagePlus imp = WindowManager.getImage(wList[i]);
            if (imp!=null)
                titles[i] = imp.getTitle();
            else
                titles[i] = "";
                                
        }
        
        if (!showDialog())
            return;
        
        
        /* some tests */
        
        if (i1.getType()!=i2.getType()) {
                    IJ.showMessage("FOCUS-SIS", "Images must have same type");
            return;
        }
        
        if (i1.getWidth()!=i2.getWidth()) {
                    IJ.showMessage("FOCUS-SIS", "Images must have same width");
            return;
        }
        
        if (i1.getHeight()!=i2.getHeight()) {
                    IJ.showMessage("FOCUS-SIS", "Images must have same height");
            return;
        }
        
        if (i1.getStackSize()!=i2.getStackSize()) {
                    IJ.showMessage("FOCUS-SIS", "Stacks must have same number of slices");
            return;
        }
        
        
        if (i1.getCalibration().calibrated())  i1=uncalibrate(i1);
        if (i2.getCalibration().calibrated())  i2=uncalibrate(i2);
        
        cMask=buildMask(i1,i2);
        
        GetNormParam(i1,i2,cMask);
                
        SubtractImages(i1, i2, cMask);
    }
    
    public boolean showDialog() {
        GenericDialog gd = new GenericDialog("FOCUS-SIS");
        gd.addChoice("Image 1 (peri-ictal):", titles, titles[0]);
        gd.addChoice("Image 2 (inter-ictal):", titles, titles[1]);
        gd.addChoice("Subtraction:", subt, subt[subtract]);
        gd.showDialog();
        if (gd.wasCanceled())
            return false;
        int i1Index = gd.getNextChoiceIndex();
        int i2Index = gd.getNextChoiceIndex();
        subtract = gd.getNextChoiceIndex();
        if (subtract==LASSEN) {
                GenericDialog ngd = new GenericDialog("LASSEN's CORRECTION");
                ngd.addChoice("Injected Molecule:", mol, mol[molecule]);
                ngd.showDialog();
                molecule = ngd.getNextChoiceIndex();
                if (ngd.wasCanceled()) return false;
        }
        if (molecule==HMPAO) alpha=1.2;       
        
        i1 = WindowManager.getImage(wList[i1Index]);
        i2 = WindowManager.getImage(wList[i2Index]);
        return true;
    }
    
    ImagePlus uncalibrate(ImagePlus imag) {
        ImageStack st1 = imag.getStack();
        int n = st1.getSize();
        ImageStack st2 = new ImageStack(st1.getWidth(),st1.getHeight());
        float[] ctable = imag.getCalibration().getCTable();
        for (int i=1; i<=n; i++) {
                ImageProcessor ipr1 = st1.getProcessor(i);
                ImageProcessor ipr2 = ipr1.duplicate(); 
                ipr2.setCalibrationTable(ctable);
                ipr2 = ipr2.convertToFloat();
                st2.addSlice(st1.getSliceLabel(i), ipr2);
        }
               
        ImagePlus imag2 =  new ImagePlus("", st2);
        return imag2;
    }
    
    
    
   // Construction of the brain mask mandatory to further calculation 
   
   public ImagePlus buildMask(ImagePlus i1, ImagePlus i2) {
        double max1=0.0, max2=0.0, v1, v2;
        String name;
        int width  = i1.getWidth();
        int height = i1.getHeight();
        ImageProcessor ip1, ip2, ipstackmask;
        int nslices = i1.getStackSize();
        ImageStack stack1 = i1.getStack();
        ImageStack stack2 = i2.getStack();
        
        for (int n=1; n<=nslices; n++) {
            ip1 = stack1.getProcessor(n);
            ip2 = stack2.getProcessor(n);
            for (int x=0; x<width; x++) {
                for (int y=0; y<height; y++) {
                    v1 = ip1.getPixelValue(x,y);
                    v2 = ip2.getPixelValue(x,y);          
                    if (v1>max1) max1=v1;
                    if (v2>max2) max2=v2;
                }
            }
        }
        
        max1/=3;max2/=3; // thresholds to build the binary mask = 1/3max(stack)
        
        ImagePlus mask = NewImage.createByteImage("MASK", width, height, nslices, NewImage.FILL_BLACK);
        ImageStack stackmask = mask.getStack();
        
        for (int n=1; n<=nslices; n++) {
            ip1 = stack1.getProcessor(n);
            ip2 = stack2.getProcessor(n);
            ipstackmask = stackmask.getProcessor(n);
            for (int x=0; x<width; x++) {
                for (int y=0; y<height; y++) {
                    v1 = ip1.getPixelValue(x,y);
                    v2 = ip2.getPixelValue(x,y);
                    if (v1>=max1 || v2>=max2) ipstackmask.putPixelValue(x, y, 255); // Brain
                    else ipstackmask.putPixelValue(x, y, 0); // Background
                }
            }
            
        }
        //mask.show();
        return mask;
   }
    
// Images normalization : according to paper mentioned above (z values) 
    
    public void GetNormParam(ImagePlus i1, ImagePlus i2, ImagePlus cMask) {
        double v1, v2, vmask, k, var1=0.0, var2=0.0, buf1=0.0, buf2=0.0;
        String name;
        int compte=0;
        double moy1,moy2,sum1=0.0,sum2=0.0;
        int width  = i1.getWidth();
        int height = i1.getHeight();
        ImageProcessor ip1, ip2, ipmask, ipzv1, ipzv2;
        int nslices = i1.getStackSize();
        ImageStack stack1 = i1.getStack();
        ImageStack stack2 = i2.getStack();
        ImageStack stackmask = cMask.getStack();
        
        for (int n=1; n<=nslices; n++) {
            ip1 = stack1.getProcessor(n);
            ip2 = stack2.getProcessor(n);
            ipmask = stackmask.getProcessor(n);
            
            for (int x=0; x<width; x++) {
                for (int y=0; y<height; y++) {
                  v1 = ip1.getPixelValue(x,y);
                  v2 = ip2.getPixelValue(x,y);
                  vmask = ipmask.getPixelValue(x,y);
                  if (vmask == 255) {
                            compte++;
                            sum1+=v1;
                            sum2+=v2;
                  }
                }
            }
        }
        MEAN1=sum1/compte;
        MEAN2=sum2/compte;
        
        name=IJ.d2s(MEAN1);
        IJ.write(" MEAN BRAIN SIGNAL in IMAGE 1");
        IJ.write(name);        
        name=IJ.d2s(MEAN2);
        IJ.write(" MEAN BRAIN SIGNAL in IMAGE 2 ");
        IJ.write(name);
        
        //  global variance  
        
        for (int n=1; n<=nslices; n++) {
            ip1 = stack1.getProcessor(n);
            ip2 = stack2.getProcessor(n);
            ipmask = stackmask.getProcessor(n);
            
            for (int x=0; x<width; x++) {
                for (int y=0; y<height; y++) {
                  v1 = ip1.getPixelValue(x,y);
                  v2 = ip2.getPixelValue(x,y);
                  vmask = ipmask.getPixelValue(x,y);
                  if (vmask == 255) {
                            buf1+=(v1-MEAN1)*(v1-MEAN1);
                            buf2+=(v2-MEAN2)*(v2-MEAN2);
                  }
                }
            }
        }
        
        var1=buf1/compte;
        var2=buf2/compte;
        var1=Math.sqrt(var1);
        var2=Math.sqrt(var2);
        
        // z-maps images 
        ImagePlus zv1 = NewImage.createFloatImage("Z1", width, height, nslices, NewImage.FILL_BLACK);
        ImageStack zv1stack = zv1.getStack();
        ImagePlus zv2 = NewImage.createFloatImage("Z2", width, height, nslices, NewImage.FILL_BLACK);
        ImageStack zv2stack = zv2.getStack();
        
        
        for (int n=1; n<=nslices; n++) {
            ip1 = stack1.getProcessor(n);
            ip2 = stack2.getProcessor(n);
            ipmask = stackmask.getProcessor(n);
            ipzv1 = zv1stack.getProcessor(n);
            ipzv2 = zv2stack.getProcessor(n);
            
            for (int x=0; x<width; x++) {
                for (int y=0; y<height; y++) {
                  v1 = ip1.getPixelValue(x,y);
                  v2 = ip2.getPixelValue(x,y);
                  vmask = ipmask.getPixelValue(x,y);
                  if (vmask == 255) {                          
                            ipzv1.putPixelValue(x, y, Math.abs( (v1-MEAN1)/var1));
                            ipzv2.putPixelValue(x, y, Math.abs( (v2-MEAN2)/var2));
                  }
                  else {
                          ipzv1.putPixelValue(x, y, 0);
                          ipzv2.putPixelValue(x, y, 0);
                  }
                }
            }
        }        
   
        // MEAN Signal in reference regions 
        
        compte=0;sum1=0.0;sum2=0.0;
        for (int n=1; n<=nslices; n++) {
            ip1 = stack1.getProcessor(n);
            ip2 = stack2.getProcessor(n);
            ipmask = stackmask.getProcessor(n);
            ipzv1 = zv1stack.getProcessor(n);
            ipzv2 = zv2stack.getProcessor(n);
            
            for (int x=0; x<width; x++) {
                for (int y=0; y<height; y++) {
                  v1 = ip1.getPixelValue(x,y);
                  v2 = ip2.getPixelValue(x,y);
                  
                  if (ipmask.getPixelValue(x,y)==255 && ipzv1.getPixelValue(x,y)<1 && ipzv2.getPixelValue(x,y)<1 ) {
                            compte++;
                            sum1+=v1;
                            sum2+=v2;
                  }
                }
            }
        }
        MEAN1=sum1/compte;
        MEAN2=sum2/compte;        
              
        name=IJ.d2s(MEAN1);
        IJ.write(" MEAN SIGNAL IN REFERENCE REGION 1 ");
        IJ.write(name);
        name=IJ.d2s(MEAN2);
        IJ.write(" MEAN SIGNAL IN REFERENCE REGION 2 ");
        IJ.write(name);
        
        return;
}    
    
    // Normalization and subtraction of input images 
    
    public void SubtractImages(ImagePlus i1, ImagePlus i2, ImagePlus cMask) {
        double v1, v2;
        int width  = i1.getWidth();
        int height = i1.getHeight();
        ImageProcessor ip1, ip2, ipmask, ipfocus, ipnorm1, ipnorm2;
        int nslices = i1.getStackSize();
        ImageStack stack1 = i1.getStack();
        ImageStack stack2 = i2.getStack();
        ImageStack stackmask = cMask.getStack();
        // Normalized images 
        ImagePlus norm1 = NewImage.createFloatImage("Image1-Norm", width, height, nslices, NewImage.FILL_BLACK);
        ImageStack norm1stack = norm1.getStack();
        
        ImagePlus norm2 = NewImage.createFloatImage("Image2-Norm", width, height, nslices, NewImage.FILL_BLACK);
        ImageStack norm2stack = norm2.getStack();
        
        
        for (int n=1; n<=nslices; n++) {
             ip1 = stack1.getProcessor(n);
             ip2 = stack2.getProcessor(n);
             ipnorm1 = norm1stack.getProcessor(n);
             ipnorm2 = norm2stack.getProcessor(n);
                                                            
             for (int x=0; x<width; x++) {
                  for (int y=0; y<height; y++) {
                      v1 = ip1.getPixelValue(x,y);
                      v2 = ip2.getPixelValue(x,y);
                      if (subtract==LASSEN) {
                         ipnorm1.putPixelValue(x,y,50*(v1/MEAN1)*alpha/ (1+alpha-v1/MEAN1));
                         ipnorm2.putPixelValue(x,y,50*(v2/MEAN2)*alpha/ (1+alpha-v2/MEAN2));
                         /* According to LASSEN NA et al. The retention of 99mTc-HMPAO in the human brain
                         after intracarotid bolus injection: a kinetic analysis. J Cereb Blood Flow Metab
                         1988; 8(6): S13-22
                         and FRIBERG L et al. Retention of 99mTc-bicisate in the human brain after 
                         intracarotid injection. J Cereb Blood Flow Metab 1994; 14 Suppl 1:S19-27.*/
                       }
                       else {
                         ipnorm1.putPixelValue(x,y,v1*MEAN2/MEAN1);
                         ipnorm2.putPixelValue(x,y,v2);
                       }
                  }
             }
        }
        
        norm1.show();
        norm2.show();
        
        // Difference image                
        ImageStack stackfocus = new ImageStack(width, height);

        for (int n=1; n<=nslices; n++) {
            ip1 = norm1stack.getProcessor(n);
            ip2 = norm2stack.getProcessor(n);
            ipmask = stackmask.getProcessor(n);
            ipfocus = ip1.duplicate();
            for (int x=0; x<width; x++) {
                for (int y=0; y<height; y++) {
                    if (ipmask.getPixelValue(x,y) == 255) {
                        v1 = ip1.getPixelValue(x,y);
                        v2 = ip2.getPixelValue(x,y);
                        if (v1>v2) {
                                if (subtract==LASSEN) ipfocus.putPixelValue(x, y, v1-v2); 
                                else ipfocus.putPixelValue(x, y, 100*(v1-v2)/v2);
                        }
                        else ipfocus.putPixelValue(x, y, 0);
                    }
                    else ipfocus.putPixelValue(x, y, 0);
                }   
            }
            stackfocus.addSlice("", ipfocus);
          }
        ImagePlus focus = new ImagePlus("FOCUSALL", stackfocus);
        focus.show();
        LutLoader lut = new LutLoader();
        lut.run("fire");
        IJ.write(" SUBTRACTION DONE ");
                       
    }
 
}

   